"""Strategies for generating torch devices."""

from __future__ import annotations

from collections.abc import Sequence

import torch
from hypothesis import strategies as st

AVAILABLE_CPU_DEVICES = [torch.device("cpu")]
AVAILABLE_CUDA_DEVICES = [torch.device("cuda", i) for i in range(torch.cuda.device_count())]
AVAILABLE_MPS_DEVICES = [torch.device("mps")] if torch.backends.mps.is_available() else []
AVAILABLE_META_DEVICES = [torch.device("meta")]

AVAILABLE_PHYSICAL_DEVICES = AVAILABLE_CPU_DEVICES + AVAILABLE_CUDA_DEVICES + AVAILABLE_MPS_DEVICES


def device_strategy(
    *,
    devices: Sequence[torch.device] | None = None,
    allow_meta_device: bool = False,
) -> st.SearchStrategy[torch.device]:
    """Strategy for generating torch devices.

    Args:
        devices: A sequence of devices to sample from. If None, all available devices are sampled.
        allow_meta_device: Whether to allow the meta device.

    Returns:
        A strategy for generating torch devices.

    """
    if devices is None:
        devices = AVAILABLE_PHYSICAL_DEVICES
    if allow_meta_device:
        devices = list(devices) + AVAILABLE_META_DEVICES
    assert devices is not None
    return st.sampled_from(devices)


st.register_type_strategy(torch.device, device_strategy())
if torch.cuda.is_available():
    st.register_type_strategy(torch.cuda.device, device_strategy(devices=AVAILABLE_CUDA_DEVICES))



"""Strategies for generating PyTorch Modules."""

from __future__ import annotations

from typing import TypeVar, Sequence, Mapping

from hypothesis_torch import inspection_util
import torch
from hypothesis import strategies as st
from torch import nn

from hypothesis_torch import utils

T = TypeVar("T")

SENSIBLE_FLOATS = st.floats(
    min_value=-10,
    max_value=10,
    allow_nan=False,
    allow_infinity=False,
    allow_subnormal=False,
)
SENSIBLE_POSITIVE_FLOATS = st.floats(
    min_value=0,
    max_value=10,
    allow_nan=False,
    allow_infinity=False,
    allow_subnormal=False,
)
POSITIVE_INTS = st.integers(min_value=1)


@st.composite
def lower_upper_strategy(draw: st.DrawFn) -> tuple[float, float]:
    """Strategy for generating a pair of floats where the first is less than the second.

    Args:
        draw: The draw function provided by `hypothesis`.

    Returns:
        A pair of floats where the first is less than the second.

    """
    lower = draw(SENSIBLE_FLOATS)
    upper = draw(SENSIBLE_FLOATS.filter(lambda x: x > lower))
    return lower, upper


@st.composite
def rrelu_strategy(draw: st.DrawFn) -> nn.RReLU:
    """Strategy for generating instances of `nn.RReLU` by drawing values for its constructor.

    Args:
        draw: The draw function provided by `hypothesis`.

    Returns:
        An instance of `nn.RReLU`.
    """
    lower, upper = draw(lower_upper_strategy())
    inplace = draw(st.booleans())
    return nn.RReLU(lower, upper, inplace)


@st.composite
def hard_tanh_strategy(draw: st.DrawFn) -> nn.Hardtanh:
    """Strategy for generating instances of `nn.Hardtanh` by drawing values for its constructor.

    Args:
        draw: The draw function provided by `hypothesis`.

    Returns:
        An instance of `nn.Hardtanh`.

    """
    min_value = draw(SENSIBLE_FLOATS)
    max_value = draw(SENSIBLE_FLOATS.filter(lambda x: x > min_value))
    inplace = draw(st.booleans())
    return nn.Hardtanh(min_value, max_value, inplace)


activation_strategies: dict[type[nn.Module], st.SearchStrategy[nn.Module]] = {
    nn.Identity: inspection_util.signature_to_strategy(nn.Identity),
    nn.ELU: inspection_util.signature_to_strategy(nn.ELU, alpha=SENSIBLE_FLOATS, inplace=st.booleans()),
    nn.Hardshrink: inspection_util.signature_to_strategy(nn.Hardshrink, lambd=SENSIBLE_FLOATS),
    nn.Hardsigmoid: inspection_util.signature_to_strategy(nn.Hardsigmoid, inplace=st.booleans()),
    nn.Hardtanh: hard_tanh_strategy(),
    nn.Hardswish: inspection_util.signature_to_strategy(nn.Hardswish, inplace=st.booleans()),
    nn.LeakyReLU: inspection_util.signature_to_strategy(
        nn.LeakyReLU, negative_slope=SENSIBLE_FLOATS, inplace=st.booleans()
    ),
    nn.LogSigmoid: inspection_util.signature_to_strategy(nn.LogSigmoid),
    # TODO: nn.MultiheadAttention, although in the `Non-linear activations` section, does not have the same shape
    #  inside and outside
    # TODO: nn.PReLU(num_parameters=1, init=0.25, device=None, dtype=None)
    # TODO: PReLU might depend on the input shape
    # TODO: num_parameters (int) – number of a to learn. Although it takes an int as input, there is only two
    #  values are legitimate: 1, or the number of channels at input. Default: 1
    nn.PReLU: inspection_util.signature_to_strategy(nn.PReLU, num_parameters=st.just(1), init=SENSIBLE_FLOATS),
    nn.ReLU: inspection_util.signature_to_strategy(nn.ReLU, inplace=st.booleans()),
    nn.ReLU6: inspection_util.signature_to_strategy(nn.ReLU6, inplace=st.booleans()),
    nn.RReLU: rrelu_strategy(),
    nn.SELU: inspection_util.signature_to_strategy(nn.SELU, inplace=st.booleans()),
    nn.CELU: inspection_util.signature_to_strategy(
        nn.CELU, alpha=SENSIBLE_FLOATS.filter(lambda x: abs(x) > 1e-5), inplace=st.booleans()
    ),
    nn.GELU: inspection_util.signature_to_strategy(nn.GELU, approximate=st.sampled_from(["none", "tanh"])),
    nn.Sigmoid: inspection_util.signature_to_strategy(nn.Sigmoid),
    nn.SiLU: inspection_util.signature_to_strategy(nn.SiLU, inplace=st.booleans()),
    nn.Mish: inspection_util.signature_to_strategy(nn.Mish, inplace=st.booleans()),
    nn.Softplus: inspection_util.signature_to_strategy(
        nn.Softplus, beta=SENSIBLE_FLOATS, threshold=SENSIBLE_POSITIVE_FLOATS
    ),
    nn.Softshrink: inspection_util.signature_to_strategy(nn.Softshrink, lambd=SENSIBLE_POSITIVE_FLOATS),
    nn.Softsign: inspection_util.signature_to_strategy(nn.Softsign),
    nn.Tanh: inspection_util.signature_to_strategy(nn.Tanh),
    nn.Tanhshrink: inspection_util.signature_to_strategy(nn.Tanhshrink),
    nn.Threshold: inspection_util.signature_to_strategy(
        nn.Threshold, threshold=SENSIBLE_FLOATS, value=SENSIBLE_FLOATS, inplace=st.booleans()
    ),
    # TODO: nn.GLU depends on the input shape
}


def same_shape_activation_strategy(
    allowed_activation_functions: Sequence[type[nn.Module]] | Sequence[st.SearchStrategy[nn.Module]] | None = None,
) -> st.SearchStrategy[nn.Module]:
    """Strategy for generating activation functions that have the same shape input and output shape.

    Args:
        allowed_activation_functions: Activation functions to sample from.
            - If a sequence of strategies is provided, only these strategies are sampled from.
            - If a sequence of module types is provided, the elements are the activation functions to sample from.
            - If `None`, all supported activation functions are sampled.
            - For `None` or sequence of module type inputs, see complete list of supported activation functions in
            `activation_strategies`.

    Returns:
        A strategy for generating activation functions that have the same shape input and output shape.
    """
    if allowed_activation_functions is None:
        allowed_activation_functions = list(activation_strategies.keys())
    assert allowed_activation_functions is not None
    strategies: list[st.SearchStrategy[nn.Module]] = []
    for activation_function in allowed_activation_functions:
        if isinstance(activation_function, st.SearchStrategy):
            strategies.append(activation_function)
        elif activation_function in activation_strategies:
            strategies.append(activation_strategies[activation_function])
        else:
            raise ValueError(f"Unsupported activation function: {activation_function}")
    return st.one_of(strategies)


@st.composite
def linear_network_strategy(
    draw: st.DrawFn,
    input_shape: tuple[int, ...] | torch.Size | st.SearchStrategy[tuple[int, ...]] | st.SearchStrategy[torch.Size],
    output_shape: tuple[int, ...] | torch.Size | st.SearchStrategy[tuple[int, ...]] | st.SearchStrategy[torch.Size],
    activation_layer: nn.Module | st.SearchStrategy[nn.Module],
    hidden_layer_size: int | st.SearchStrategy[int],
    num_hidden_layers: int | st.SearchStrategy[int],
    device: torch.device | st.SearchStrategy[torch.device],
) -> nn.Module:
    """Strategy for generating random Torch sequential networks of linear layers with activation functions.

    Args:
        draw: The draw function provided by `hypothesis`.
        input_shape: The shape of the input tensor. If a strategy is provided, it will be drawn from.
        output_shape: The shape of the output tensor. If a strategy is provided, it will be drawn from.
        activation_layer: Activation layer to use. If a strategy is provided, it will be drawn from.
        hidden_layer_size: The size of the hidden layers. If a strategy is provided, it will be drawn from.
        num_hidden_layers: The maximum depth of the network. If a strategy is provided, it will be drawn from.
        device: The device on which to place the network. If a strategy is provided, it will be drawn from.

    Returns:
        A strategy for generating random linear networks.

    """
    if isinstance(input_shape, st.SearchStrategy):
        input_shape = draw(input_shape)
    if len(input_shape) not in {1, 2}:
        raise ValueError(f"input_shape must be a 1D or 2D tuple, but got {input_shape}")
    *_, input_size = input_shape

    if isinstance(output_shape, st.SearchStrategy):
        output_shape = draw(output_shape)
    if len(output_shape) not in {1, 2}:
        raise ValueError(f"output_shape must be a 1D or 2D tuple, but got {output_shape}")
    *_, output_size = output_shape

    if isinstance(device, st.SearchStrategy):
        device = draw(device)

    if not isinstance(activation_layer, st.SearchStrategy):
        activation_layer_strategy = st.just(activation_layer)
    else:
        activation_layer_strategy = activation_layer

    # Several activation functions are not supported on MPS devices
    if device.type == "mps":
        unsupported_mps_activations = [nn.Hardshrink, nn.RReLU]

        activation_layer_strategy = activation_layer_strategy.filter(
            lambda x: not isinstance(x, tuple(unsupported_mps_activations))
        )

    if not isinstance(hidden_layer_size, st.SearchStrategy):
        hidden_layer_size = st.just(hidden_layer_size)
    if isinstance(num_hidden_layers, st.SearchStrategy):
        num_hidden_layers = draw(num_hidden_layers)

    with device:
        interior_layer_sizes = draw(
            st.lists(
                hidden_layer_size,
                min_size=num_hidden_layers,
                max_size=num_hidden_layers,
            )
        )
        layer_sizes = [input_size, *interior_layer_sizes, output_size]
        layers: list[nn.Module] = [nn.Linear(a, b) for a, b in utils.pairwise(layer_sizes)]
        activations: list[nn.Module] = draw(
            st.lists(activation_layer_strategy, min_size=len(layers), max_size=len(layers)),
        )
        return nn.Sequential(*utils.alternate(layers, activations)).to(device)


def convolution_output_shape(
    input_shape: tuple[int, ...] | torch.Tensor,
    kernel_size: tuple[int, ...] | torch.Tensor,
    stride: tuple[int, ...] | torch.Tensor,
    padding: tuple[int, ...] | torch.Tensor,
    dilation: tuple[int, ...] | torch.Tensor,
) -> tuple[int, ...]:
    """Calculate the output shape of a convolutional layer."""
    input_shape = torch.tensor(input_shape, dtype=torch.uint8)
    kernel_size = torch.tensor(kernel_size, dtype=torch.uint8)
    stride = torch.tensor(stride, dtype=torch.uint8)
    padding = torch.tensor(padding, dtype=torch.uint8)
    dilation = torch.tensor(dilation, dtype=torch.uint8)
    return tuple(torch.floor((input_shape + 2 * padding - dilation * (kernel_size - 1) - 1) / stride + 1).tolist())



"""Hypothesis strategies for various Pytorch structures (including tensors and modules).

[Hypothesis](https://hypothesis.readthedocs.io/en/latest/) is a powerful property-based testing library for Python. It
lacks built-in support for Pytorch tensors and modules, so this library provides strategies for generating them.
"""

__version__ = "0.6.0"
import importlib.util

from hypothesis_torch.device import (
    device_strategy,
    AVAILABLE_CPU_DEVICES,
    AVAILABLE_CUDA_DEVICES,
    AVAILABLE_MPS_DEVICES,
    AVAILABLE_META_DEVICES,
    AVAILABLE_PHYSICAL_DEVICES,
)
from hypothesis_torch.dtype import (
    dtype_strategy,
    FLOAT_DTYPES,
    INT_DTYPES,
    SIGNED_INT_DTYPES,
    UNSIGNED_INT_DTYPES,
    COMPLEX_DTYPES,
    BOOL_DTYPES,
    ALL_DTYPES,
)
from hypothesis_torch.layout import layout_strategy
from hypothesis_torch.memory_format import memory_format_strategy
from hypothesis_torch.module import linear_network_strategy, same_shape_activation_strategy
from hypothesis_torch.optim import optimizer_strategy, optimizer_type_strategy, OptimizerConstructorWithOnlyParameters
from hypothesis_torch.register_random_torch_state import TORCH_RANDOM_WRAPPER
from hypothesis_torch.tensor import tensor_strategy


if importlib.util.find_spec("transformers") is not None:
    # Import Hugging Face strategies if transformers is installed
    from hypothesis_torch.huggingface import transformer_strategy, OFFICIALLY_SUPPORTED_TRANSFORMERS



"""Strategies for generating PyTorch tensors."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any, Sequence

from hypothesis.internal.floats import float_of
from typing_extensions import Final

import hypothesis.extra.numpy as numpy_st
import torch
from hypothesis import strategies as st, reject
from hypothesis.strategies._internal import numbers as st_numbers
import hypothesis

import hypothesis_torch
from hypothesis_torch import dtype as dtype_module

_NOT_MPS_DEVICES: Final[Sequence[torch.device]] = (
    hypothesis_torch.AVAILABLE_CPU_DEVICES + hypothesis_torch.AVAILABLE_CUDA_DEVICES
)
"""All devices that are not MPS devices (since MPS devices do not yet have full torch support)."""

_ALLOWED_DEVICES_FROM_DTYPE: Final[Mapping[torch.dtype, Sequence[torch.device]]] = {
    torch.bool: hypothesis_torch.AVAILABLE_PHYSICAL_DEVICES,
    torch.uint8: hypothesis_torch.AVAILABLE_PHYSICAL_DEVICES,
    torch.int8: hypothesis_torch.AVAILABLE_PHYSICAL_DEVICES,
    torch.int16: hypothesis_torch.AVAILABLE_PHYSICAL_DEVICES,
    torch.int32: hypothesis_torch.AVAILABLE_PHYSICAL_DEVICES,
    torch.int64: hypothesis_torch.AVAILABLE_PHYSICAL_DEVICES,
    torch.float16: hypothesis_torch.AVAILABLE_PHYSICAL_DEVICES,
    torch.float32: hypothesis_torch.AVAILABLE_PHYSICAL_DEVICES,
    # MPS devices do not support tensors with dtype torch.float64 and bfloat16
    torch.float64: _NOT_MPS_DEVICES,
    torch.bfloat16: _NOT_MPS_DEVICES,
    torch.complex64: hypothesis_torch.AVAILABLE_PHYSICAL_DEVICES,
    torch.complex128: hypothesis_torch.AVAILABLE_PHYSICAL_DEVICES,
}
"""A mapping from dtype to the devices that support that dtype."""


@st.composite
def tensor_strategy(
    draw: st.DrawFn,
    dtype: torch.dtype | st.SearchStrategy[torch.dtype],
    shape: int | st.SearchStrategy[int] | tuple[int, ...] | st.SearchStrategy[tuple[int, ...]],
    *,
    elements: st.SearchStrategy[Any] | Mapping[str, Any] | None = None,
    fill: st.SearchStrategy[Any] | None = None,
    unique: bool | st.SearchStrategy[bool] = False,
    device: torch.device | st.SearchStrategy[torch.device] | None = None,
    requires_grad: bool | st.SearchStrategy[bool] | None = None,
    pin_memory: bool | st.SearchStrategy[bool] | None = None,
    layout: torch.layout | st.SearchStrategy[torch.layout] | None = None,
    memory_format: torch.memory_format | st.SearchStrategy[torch.memory_format] | None = None,
) -> torch.Tensor:
    """A strategy for generating PyTorch tensors.

    Args:
        draw: The draw function provided by `hypothesis`.
        dtype: Any PyTorch dtype or a strategy for generating PyTorch dtypes.
        shape: The shape of the tensor. Can be an integer >= 0, a tuple of such integers, or a strategy for generating
            such values.
        elements: A strategy for generating elements of the tensor. If `None`, a suitable strategy is inferred from
            the dtype. Note that this may give any legal value (including NaNs and infinities for floats).
        fill: A strategy for generating a single background value for the tensor. If None, a suitable default will
            be inferred based on the other arguments. If set to `~hypothesis.strategies.nothing` then filling
            behaviour will be disabled entirely and every element will be generated independently.
        unique: Whether the tensor's elements should all be distinct from one another. Note that multiple NaN values
            may still be allowed.
        device: The device on which to place the tensor. If None, the default device is used.
        requires_grad: Whether the tensor requires gradients. If None, a suitable default will be inferred based on
            the other arguments.
        pin_memory: Whether the tensor should be pinned in memory. If None, a suitable default will be inferred based
            on the other arguments.
        layout: The memory layout of the tensor. If None, a suitable default will be inferred based on the other
            arguments. Note that sparse layouts are not supported on MPS devices.
        memory_format: The memory format of the tensor. If None, a suitable default will be inferred based on the other
            arguments. Note that channel_last memory formats are only supported for 4D tensors and channel_last_3d
            memory formats are only supported for 5D tensors.

    Returns:
        A strategy for generating PyTorch tensors.

    """
    # We will pre-sample the dtype so that we can cast it to a concrete numpy dtype
    if isinstance(dtype, st.SearchStrategy):
        dtype = draw(dtype)
    numpy_dtype = dtype_module.numpy_dtype_map[dtype]

    # We will pre-sample the device so that we can cast it to a concrete torch device
    if device is None:
        device = st.sampled_from(_ALLOWED_DEVICES_FROM_DTYPE[dtype])
    if isinstance(device, st.SearchStrategy):
        device = draw(device)
    # MPS devices do not support tensors with dtype torch.float64 and bfloat16
    hypothesis.assume(device is None or device.type != "mps" or dtype not in (torch.float64, torch.bfloat16))

    if layout is None:
        layout = st.from_type(torch.layout)
    if isinstance(layout, st.SearchStrategy):
        layout = draw(layout)
    # MPS devices do not support sparse tensors
    hypothesis.assume(device is None or device.type != "mps" or layout != torch.sparse_coo)

    # If the dtype is an integer, we need to make sure that the elements are integers within the dtype's range
    if dtype in dtype_module.INT_DTYPES and isinstance(elements, st.SearchStrategy):
        info = torch.iinfo(dtype)
        elements = elements.filter(lambda x: info.min <= x <= info.max)

    # If the dtype is a float, then we need to make sure that only elements that can be represented exactly are
    # generated
    if dtype in dtype_module.FLOAT_DTYPES and elements is not None:
        if dtype == torch.bfloat16:
            # Since we do not directly support bfloat16 in numpy, we will generate float32.
            # This still means that we will occasionally generate values that exceed the max/min of bfloat16.
            # All other values (within the range) will be simply truncated below when casting the numpy array to
            # a bfloat16 tensor.
            bfloat16_info = torch.finfo(torch.bfloat16)
            elements = elements.filter(lambda x: bfloat16_info.min <= x <= bfloat16_info.max)

        width = dtype_module.float_width_map[dtype]
        if width < 64:

            def downcast(x: float) -> float:
                """Downcast a float to a smaller width.

                This function is used to ensure that only floats that can be represented exactly are generated.

                Adapted from `hypothesis.strategies.numbers.floats`.

                Args:
                    x: The float to downcast.

                Returns:
                    The downcasted float.
                """
                try:
                    return hypothesis.internal.floats.float_of(x, width)
                except OverflowError:  # pragma: no cover
                    hypothesis.reject()

            elements = elements.map(downcast)

    if isinstance(unique, st.SearchStrategy):
        unique = draw(unique)

    ndarray_strategy = numpy_st.arrays(numpy_dtype, shape, elements=elements, fill=fill, unique=unique)
    tensor = draw(ndarray_strategy.map(torch.from_numpy))

    if pin_memory is None:
        pin_memory = st.booleans() if device.type == "cuda" else False
    if isinstance(pin_memory, st.SearchStrategy):
        pin_memory = draw(pin_memory)
    if pin_memory and device.type == "cuda":
        tensor = tensor.pin_memory()

    tensor = tensor.to(device=device, dtype=dtype)

    if requires_grad is None:
        requires_grad = st.booleans() if dtype in dtype_module.FLOAT_DTYPES else False
    if isinstance(requires_grad, st.SearchStrategy):
        requires_grad = draw(requires_grad)
    tensor.requires_grad_(requires_grad)

    if layout == torch.strided:
        tensor = tensor.contiguous()
    elif layout == torch.sparse_coo:
        # TODO: Implement coalesced handling
        tensor = tensor.to_sparse_coo()
    else:
        raise ValueError(f"Unsupported layout: {layout}")

    # MEMORY FORMAT HANDLING
    if memory_format is None:
        permitted_memory_formats = [torch.contiguous_format]
        if len(tensor.shape) == 4:
            permitted_memory_formats.append(torch.channels_last)
        if len(tensor.shape) == 5:
            permitted_memory_formats.append(torch.channels_last_3d)
        memory_format = st.sampled_from(permitted_memory_formats)
    if isinstance(memory_format, st.SearchStrategy):
        memory_format = draw(memory_format)
    # Filter out memory formats that are not supported by the tensor's shape, in case a user-specified strategy is used.
    # channel_last memory format is only supported for 4D tensors
    hypothesis.assume(memory_format != torch.channels_last or len(tensor.shape) == 4)
    # channel_last_3d memory format is only supported for 5D tensors
    hypothesis.assume(memory_format != torch.channels_last_3d or len(tensor.shape) == 5)
    tensor = tensor.to(memory_format=memory_format)

    return tensor


st.register_type_strategy(
    torch.Tensor,
    tensor_strategy(
        dtype=dtype_module.dtype_strategy(),
        shape=st.integers(min_value=0, max_value=10),  # Discourage large tensors
    ),
)



"""Strategies for generating PyTorch dtypes."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Mapping, Literal

from typing_extensions import Final

import numpy as np
import numpy.typing as npt
import torch
from hypothesis import strategies as st

SIGNED_INT_DTYPES: Final[tuple[torch.dtype, ...]] = (
    torch.int8,
    torch.int16,
    torch.int32,
    torch.int64,
)
"""All signed integer dtypes supported by PyTorch."""
UNSIGNED_INT_DTYPES: Final[tuple[torch.dtype, ...]] = (torch.uint8,)
"""All unsigned integer dtypes supported by PyTorch."""
INT_DTYPES: Final[tuple[torch.dtype, ...]] = SIGNED_INT_DTYPES + UNSIGNED_INT_DTYPES
"""All integer (both signed and unsigned) dtypes supported by PyTorch."""
FLOAT_DTYPES: Final[tuple[torch.dtype, ...]] = (
    torch.bfloat16,
    torch.float16,
    torch.float32,
    torch.float64,
)
"""All floating point dtypes supported by PyTorch."""
COMPLEX_DTYPES: Final[tuple[torch.dtype, ...]] = (
    torch.complex64,
    torch.complex128,
)
"""All complex dtypes supported by PyTorch."""
BOOL_DTYPES: Final[tuple[torch.dtype, ...]] = (torch.bool,)
"""All boolean dtypes supported by PyTorch."""

ALL_DTYPES: Final[tuple[torch.dtype, ...]] = INT_DTYPES + FLOAT_DTYPES + COMPLEX_DTYPES + BOOL_DTYPES
"""All dtypes supported by PyTorch."""

numpy_dtype_map: Final[Mapping[torch.dtype, npt.DTypeLike]] = {
    torch.int8: np.int8,
    torch.int16: np.int16,
    torch.int32: np.int32,
    torch.int64: np.int64,
    torch.uint8: np.uint8,
    torch.float16: np.float16,
    torch.float32: np.float32,
    torch.float64: np.float64,
    torch.bfloat16: np.float32,  # Numpy does not have a bf16, but it is a strict subset of fp32
    torch.complex64: complex,
    torch.complex128: complex,
    torch.bool: np.bool_,
}
"""A mapping from torch dtypes to numpy dtypes. Useful for generating tensors of arbitrary dtypes from the builtin
numpy strategies."""

float_width_map: Final[Mapping[torch.dtype, Literal[16, 32, 64]]] = {
    torch.float16: 16,
    torch.bfloat16: 32,  # Numpy does not have a bf16, but it is a strict subset of f32
    torch.float32: 32,
    torch.float64: 64,
}

assert set(numpy_dtype_map.keys()) == set(ALL_DTYPES)
assert set(float_width_map.keys()) == set(FLOAT_DTYPES)


def dtype_strategy(dtypes: Sequence[torch.dtype] | None = None) -> st.SearchStrategy[torch.dtype]:
    """Strategy for generating torch dtypes.

    Args:
        dtypes: A strategy for generating elements of the dtype. If `None`, all dtypes are sampled.

    Returns:
        A strategy for generating torch dtypes.
    """
    if dtypes is None:
        dtypes = ALL_DTYPES
    return st.sampled_from(dtypes)


st.register_type_strategy(torch.dtype, dtype_strategy())
